/*
 *	Asteroids3D - a first person game of blowing up asteroids
 *	Copyright (C) 2000 Stuart Mark Pomerantz <smp [at] psc edu>
 *	Copyright © Jan Engelhardt <jengelh [at] medozas de>, 2003 - 2005
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *	Stuart Pomerantz
 *	3935 Stonecliffe drive
 *	Monroeville, PA 15146
 */
#include <GL/glut.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "asteroids3D.h"
#include "vecops.h"
#ifndef S_IRUGO
#	define S_IRUGO (S_IRUSR | S_IRGRP | S_IROTH)
#endif
#ifndef S_IWUGO
#	define S_IWUGO (S_IWUSR | S_IWGRP | S_IWOTH)
#endif

/* This stuff is for the score file */
#define MAX_FIELD_LEN             64
#define MAX_LINE_LEN              256
#define SCORE_FILE                "/tmp/asteroids-scores"
#define NUM_SCORES                10
#define SCORE_FILE_SEPARATOR_CHAR ','
#define SCORE_FILE_COMMENT_CHAR   '#'
#define SCORE_FILE_ESCAPE_CHAR    '\\'

struct score_file_entry {
	char name[MAX_FIELD_LEN], date[MAX_FIELD_LEN];
	unsigned int score;
	int new_score_flag; /* this value isn't saved to disk */
};

static int compare_scores(const void *, const void *);
static void read_score_file(void);
static void score_display(void);
static void score_keyboard_handler(unsigned char, int, int);
static void write_score_file(void);

/* last but not least, what would a game be without a score */
unsigned int score = 0;
double difficulty_multiplier = 1;

/* Data structure to hold the scores read from the score file */
static struct score_file_entry score_entry[NUM_SCORES] = {};

/*---------------------------------------------------------------------------*/
void add_rock_to_score(const struct asteroid *a)
{
	/* Scoring works like this:
	 *
	 * - small rocks are worth more than big ones
	 * - fast rocks are worth more than slow ones
	 *
	 * so I am doing (100/size)*speed <=> (100 * speed / size)
	 *
	 * Adding one below so I do not inadvertantly
	 * score a zero for a stationary or slow asertoid.
	 */
	score += 100.0 * (vec_length(&a->velocity) + 1) / (a->type + 1.0);
}

static void read_score_file(void)
{
	char buf[MAX_LINE_LEN];
	int i = 0;
	FILE *fp;

	if ((fp = fopen(SCORE_FILE, "r")) == NULL)
		return;

	while (i < NUM_SCORES && fgets(buf, sizeof(buf), fp) != NULL) {
		char *bp = buf, *res;
		struct score_file_entry *se;

		strip_nl(bp);
		if (*bp == '#' || *bp == '\0')
			continue;

		se = &score_entry[i++];

		if ((res = strsep(&bp, ",")) != NULL) {
			strncpy(se->name, res, sizeof(se->name) - 1);
			se->name[sizeof(se->name) - 1] = '\0';
		}
		if ((res = strsep(&bp, ",")) != NULL)
			se->score = strtol(res, NULL, 0);
		if (bp != NULL) {
			strncpy(se->date, bp, sizeof(se->date) - 1);
			se->date[sizeof(se->date) - 1] = '\0';
		}

		se->new_score_flag = 0;
	}

	fclose(fp);
}

static void write_score_file(void)
{
	FILE *fp;
	int i;

	if ((fp = fopen(SCORE_FILE, "w")) == NULL) {
		fprintf(stderr, "Could not open %s for writing: %s\n",
		        SCORE_FILE, strerror(errno));
		return;
	}

	fprintf(fp,
		"#\n"
		"# asteroids score file\n"
		"#\n"
		"# This file is automatically generated and will be "
		"recreated if deleted.\n"
		"#\n"
		"# name,score,date\n"
	);

	for (i = 0; i < NUM_SCORES; ++i) {
		struct score_file_entry *se = &score_entry[i];
		if (*se->name != '\0' && *se->date != '\0')
			fprintf(fp, "%s,%u,%s\n", se->name,
			        se->score, se->date);
	}

	fclose(fp);
	chmod(SCORE_FILE, S_IRUGO | S_IWUGO);
}

static int compare_scores(const void *pa, const void *pb)
{
	const struct score_file_entry *a = pa, *b = pb;
	return b->score - a->score;
}

void init_score_display(void)
{
	struct score_file_entry *se;
	time_t now;
	char *user;

	/* The player's score is scaled to reflect the difficulty of the game
	starting with more asteroids gets a higher score than starting with fewer.
	NUM_ASTEROIDS asteroids is the "base" score. The difficulty multiplier is
	already calculated at game start so that we can display the correct score
	already in-game. */
	score *= difficulty_multiplier;

	read_score_file();
	glutDisplayFunc(score_display);
	glutKeyboardFunc(score_keyboard_handler);

	if (score <= score_entry[NUM_SCORES-1].score)
		return;

	se = &score_entry[NUM_SCORES-1];

	if ((user = getenv("USER")) != NULL)
		strncpy(se->name, user, sizeof(se->name) - 1);
	else
		strncpy(se->name, "(unknown)", sizeof(se->name) - 1);

	se->name[sizeof(se->name) - 1] = '\0';
	se->score = score;
	se->new_score_flag = 1;
	now = time(NULL);
	strftime(se->date, sizeof(se->date), "%B %d %Y, %H:%M",
	         localtime(&now));

	qsort(score_entry, NUM_SCORES, sizeof(*score_entry), compare_scores);
	write_score_file();
}

static void score_keyboard_handler(unsigned char key, int x, int y)
{
	if (key == keymap.end_game)
		exit(EXIT_SUCCESS);
}

static void score_display(void)
{
	char buf[MAX_LINE_LEN];
	double r, g, b, a;
	int i;

	/* do some error checking for each frame. */
	glutReportErrors();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	draw_text_in_color(-0.2, 0.9, "Top Ten Scores", 0, 0, 1, 1);

	for (i = 0; i < NUM_SCORES; ++i) {
		struct score_file_entry *se = &score_entry[i];

		if (*se->name == '\0' || *se->date == '\0')
			continue;

		if (!se->new_score_flag) {
			r = g = b = a = 1;
		} else {
			r = a = 1;
			g = b = 0;
		}

		draw_text_in_color(-0.7, 0.7 - static_cast(double, i) /
		                   NUM_SCORES, se->name, r, g, b, a);

		snprintf(buf, sizeof(buf), "%u", se->score);
		draw_text_in_color(-0.2, 0.7 - static_cast(double, i) /
		                   NUM_SCORES, buf, r, g, b, a);

		draw_text_in_color(0.2, 0.7 - static_cast(double, i) /
		                   NUM_SCORES, se->date, r, g, b, a);
	}

	key_char2str(keymap.end_game, buf, sizeof(buf));
	strcat(buf, " key exits");
	draw_text_in_color(-0.2, -0.7, buf, 0.5, 0.5, 0.5, 1);
	glutSwapBuffers();
}
